/**********************************************************************

	Command:  FNCURVE  -- A function curve generation program

	Programmed by Kang, Hsin-Min.
	Primitive edition at Jun 14, 1997
	Revised at Jun 20, 1997
	Revised at Sep 13, 1997 (Add [Help] button, Change layout)
	Revised at Apr 22, 1998 Include Japenese Text in SJIS codes
				Zoom to curve extent not drawing extent

	Copyright (c) 1997, TCAM Development House, Taipei

	Status: Standard Command as well as a programming Example.

	Note:	In this primitive version, only a 2-D planear curve
		defined by parametric function in the form as P(X(t),Y(t))
		is supported and is realized using either line or arc
		segments. The curve is assumed continueous and well defined
		within the given parametric interval. No further checking or
		analysis will be done in this version.

***********************************************************************/
char	xt[80]="t";     // X(t) definition string
char	yt[80]="";      // Y(t) definition string
char	ts[20]="0";     // T0 Expression
char	te[20]="0";     // T1 Expression
char	dt[20]="1";     // dT Expression
#define GENRLINE	0
#define GENRARC 	1
#define GENRPOINT	2
char	szType[][10]={"Lines","Arcs","Pnts"};
int	GenrType=GENRARC;	// Default: Using Arc to approximate
long	Lcnt=0;
//
char	szText[6][40];	// Our display text buffer, see SetMsgText()
			// at the end of this file
COMMAND Fncurve()
{
	if ( DoDialog() <= 0 )	// Get Function Definitions
		exit();
	eval("v0=("+ts+")");
	eval("v1=("+te+")");
	eval("v2=("+dt+")");
	printf("\nParametric Curve Function: ");
	printf("\nX(t) = %s",xt);
	printf("\nY(t) = %s",yt);
	printf("\nT0 = %10.5f, evaluated by '%s'",v0,ts);
	printf("\nT1 = %10.5f, evaluated by '%s'",v1,te);
	printf("\ndT = %10.5f, evaluated by '%s'",v2,dt);
	v2 = abs(v2);
	if ( v2 == 0 )
	    exit();
	if ( v0 > v1 )
	    swap(&v0,&v1);
	switch(GenrType) {
	  case GENRLINE: DoLineSeg();	break;
	  case GENRARC:  DoCSpline();	break;
	  default:	 DoPoints();	break;
	}

	if ( FindExtend() ) {	// Modified at Apr 22, 1998
		p1 = (l1.ps+l1.pe)/2;
		L1 = (L1-p1)*1.2+p1;	// Make a larger view extent
		command("zw l1.ps l1.pe ");
	}
	else
		command("ze");  // So we can see the result on the screen
	exit();
}
//
//	Generate the curve with line segments joining each data point.
//
int	DoLineSeg()
{
	s1 = "{ double t; t=v0;";
	s2 = "p2.x="+xt+";";
	s3 = "p2.y="+yt+";";
	s4 = "for(t=v0;t<=v1;t+=v2) { ";
	s5 = "p1.x="+xt+";";
	s6 = "p1.y="+yt+";";
	s7 = "newent(L(p2,p1)); p2=p1; } }";
	return command(s1,s2,s3,s4,s5,s6,s7);
}
//
//	Generate the curve by curve-fitting the data point using cubic
//	spline. The spline is realized with dual-arc approximation on
//	each interval and is created as a POLYLINE.
//
//	The actual dual-arc results will be affected by the current
//	setting value of the system variable @SARCTYPE.
//
int	DoCSpline()
{
	long	lcnt;
	char	cnt[10];
	int	i, iseg;
	lcnt = (v1-v0)/v2;
	lcnt++;
	if ( lcnt >= 4096 )	// Limited by 64 Kb array size
		return DoLineSeg();
	iseg = @SPLINESEG;
	@SPLINESEG = 0; 	// Minimum
	cnt = prints("%ld",lcnt);
	s1 = "{ double t; POINT pt["+cnt+"];i0=0;";
	s2 = "for(t=v0;t<=v1;t+=v2) { ";
	s3 = "p1.x="+xt+";";
	s4 = "p1.y="+yt+";";
	s5 = "pt[i0++]=p1; } cspline(pt,i0,(pt[0]==pt[i0-1])?5:0);}";
	i = command(s1,s2,s3,s4,s5);
	@SPLINESEG = iseg;
	return i;
}
//
//	Generate the curve with points.
//
int	DoPoints()
{
	s1 = "{ double t; t=v0;";
	s2 = "for(t=v0;t<=v1;t+=v2) { ";
	s3 = "p1.x="+xt+";";
	s4 = "p1.y="+yt+";";
	s5 = "newent(p1); } }";
	return command(s1,s2,s3,s4,s5);
}
//
//	Variable for global usage
//
int	widMain;	// Main Dialog Window ID
int	widVport;	// Sub Window for text view port
int	nMain[6]={2,23,9,72,0x3271,0x8000};
int	nVport[5]={8,19,2,61,0x32f7};
int	nTatrb=0x70;
int	nSatrb=0x270;
int	bXt;		// X(t) Field
int	bYt;		// Y(t) Field
int	bT0;		// t0 Field
int	bT1;		// t1 Field
int	bDt;		// dt Field
int	bLcnt;		// Seg # Field
int	bOK;		// Create Button
int	bCancel;	// Quit Button
int	bPreview;	// Preview Button
int	bType;		// Generation Type
int	bHelp;		// Help button
int	IsWin;		// Is Windows Version?

int	DoDialog()
{
	int	nBtn;
	int	nRet;

	SetMsgText();

	IsWin = 1;
	switch(1) {  // Test if we are in TwinCAD ...
	  case !identify("os_type"):
	  case (os_type()<0):	// Must be in DOS version!
		IsWin = 0;
		break;
	}
	InitialData();
	OpenDialog();
	UpdateDialog();
	nRet = 0;
	for(;;) {
		nBtn = wnd_gbtn(0);
		switch(nBtn) {
		  case -3:	// ESC, Right Btn, Mid Btn, [-]
		  case -2:	// All will quit the operations
		  case -1:
		  case	0:
		  case bCancel: break;
		  case bXt:	GetXt();	continue;
		  case bYt:	GetYt();	continue;
		  case bT0:	GetT0();	continue;
		  case bT1:	GetT1();	continue;
		  case bDt:	GetDt();	continue;
		  case bType:	GetType();	continue;
		  case bPreview: DoPreview();	continue;
		  case bHelp:	show_help("","FNCURVE");        continue;
		  case bOK:
			if ( !DataOk() ) {
				beep();
				continue;
			}
			nRet =	1;
			break;
		  default:
			continue;
		}
		break;
	}
	SaveData();
	CloseDialog();
	return nRet;
}
//
//	Read previous definitions from priviate SET file.
//	No error checking is provided, since the user can
//	detect the errors and correct them easily.
//
int	InitialData()
{
	int	hdl;
	char	fname[80]="FNCURVE.SET";
	if ( identify("get_syspath") )
		fname = fullname(fname,get_syspath(IsWin?1:0));
	if ( !(hdl=fopen(fname,"r")) )
		return 0;
	fgets(xt,80,hdl);
	fgets(yt,80,hdl);
	fgets(ts,20,hdl);
	fgets(te,20,hdl);
	fgets(dt,20,hdl);
	fclose(hdl);
	return 1;
}
//
//	Save Current definitions to SET file.
//
int	SaveData()
{
	int	hdl;
	char	fname[80]="FNCURVE.SET";
	if ( identify("get_syspath") )
		fname = fullname(fname,get_syspath(IsWin?1:0));
	if ( !(hdl=fopen(fname,"w")) )
		return 0;
	fputs(xt+"\r\n",hdl);
	fputs(yt+"\r\n",hdl);
	fputs(ts+"\r\n",hdl);
	fputs(te+"\r\n",hdl);
	fputs(dt+"\r\n",hdl);
	fclose(hdl);
	return 1;
}

void	GetXt()
{
	xt = wnd_fgets(bXt,54,0x44e,xt);
	UpdateDialog();
}

void	GetYt()
{
	yt = wnd_fgets(bYt,54,0x44e,yt);
	UpdateDialog();
}

void	GetT0()
{
	ts = wnd_fgets(bT0,12,0x44e,ts);
	UpdateDialog();
}

void	GetT1()
{
	te = wnd_fgets(bT1,12,0x44e,te);
	UpdateDialog();
}

void	GetDt()
{
	dt = wnd_fgets(bDt,12,0x44e,dt);
	UpdateDialog();
}

void	GetType()
{
	if ( Lcnt ) {
	    switch(GenrType) {	    // Reset count to Line
	      case GENRARC:  Lcnt /= 2;     break;
	      case GENRPOINT: Lcnt-=1;	    break;
	    }
	}
	GenrType = (GenrType+1)%3;	// Advance type
	if ( Lcnt ) {
	    switch(GenrType) {
	      case GENRARC:  Lcnt *= 2;     break;
	      case GENRPOINT: Lcnt+=1;	    break;
	    }
	}
	UpdateDialog();
}
//
//	Check to see if the given data has no apparent error.
//
int	DataOk()
{
	if ( xt == "" ) return 0;
	if ( yt == "" ) return 0;
	if ( dt == "" ) return 0;
	if ( ts == "" ) return 0;
	if ( te == "" ) return 0;
	eval("v0=("+ts+")");
	eval("v1=("+te+")");
	eval("v2=("+dt+")");
	if ( v2 == 0 )	return 0;
	if ( v0 == v1 ) return 0;
	return 1;	// Could be ok.
}
//
//	Find the drawing extent of the function curve at the given
//	parameter range by probing a limited number of points.
//	The resulting extent is returned in L1.
//
int	FindExtend()
{
	if ( !DataOk() ) return 0;
	if ( v0 > v1 )
	    swap(&v0,&v1);
	v3 = (v1-v0)/30.;
	L1 = P(99.0e99,99.0e99),P(-99.0e99,-99.0e99);
	s1 = "{ double t; t=v0;";
	s2 = "for(t=v0;t<=v1;t+=v3) { ";
	s3 = "p1.x="+xt+";";
	s4 = "p1.y="+yt+";";
	s5 = "L1 = gextent(p1,L1); } }";
	command(s1,s2,s3,s4,s5);
	if ( L1.ps == P(99.0e99,99.0e99) ) return 0;
	if ( L1.pe == P(-99.0e99,-99.0e99) ) return 0;
	return 1;
}
//
//	Plot the function curve on the preview window.
//
int	DoPreview()
{
	int	pdmode;

	if ( !FindExtend() ) {
		beep();
		return 0;
	}

	wnd_select(widVport);
	p1 = (l1.ps+l1.pe)/2;
	L1 = (L1-p1)*1.2+p1;	// Make a larger view extent
	wnd_vport(0,L1.ps,L1.pe,1);

	setplot(7,0,0,0);	// Plot reference axis
	plotent(L(P(L1.ps.x,0),P(L1.pe.x,0)),
		L(P(0,L1.ps.y),P(0,L1.pe.y)));

	setplot(1,0,0,0);
	v2 = abs(v2);
	if ( v0 > v1 )
	    swap(&v0,&v1);
	Lcnt = (v1-v0)/v2;
	if ( GenrType == GENRPOINT ) {
		s1 = "{ double t; t=v0;";
		s2 = "for(t=v0;t<=v1;t+=v2) { ";
		s3 = "p1.x="+xt+";";
		s4 = "p1.y="+yt+";";
		s5 = "plotent(p1); } }";
		pdmode = @pdmode;
		@pdmode = 0;		// Single dot
		command(s1,s2,s3,s4,s5);
		@pdmode = pdmode;
	}
	else {
		s1 = "{ double t; t=v0;";
		s2 = "p2.x="+xt+";";
		s3 = "p2.y="+yt+";";
		s4 = "for(t=v0;t<=v1;t+=v2) { ";
		s5 = "p1.x="+xt+";";
		s6 = "p1.y="+yt+";";
		s7 = "plotent(L(p2,p1)); p2=p1; } }";
		command(s1,s2,s3,s4,s5,s6,s7);
	}
	wnd_select(widMain);
	switch(GenrType) {
	  case GENRARC: Lcnt *= 2;	break;
	  case GENRPOINT: Lcnt+=1;	break;
	}
	wnd_ftext(bLcnt,prints("%5ld",Lcnt),0xf0);
	return 1;
}
//
//	OpenDialog() -- Open main dialog window, create screen items.
//
void	OpenDialog()
{
	int	rBase, cBase;
	int	rInc=0x0801;
	widMain = wndopen(nMain,szText[0]/*"Function Curve Setup Dialog"*/,0x3f1f,nMain[5]);
	rBase = 0x0800;
	cBase = 2;
	bXt = WndEntry(rBase+rInc*0,cBase,"X(t): ",0x136);
	bYt = WndEntry(rBase+rInc*1,cBase,"Y(t): ",0x136);
	bT0 = WndEntry(rBase+rInc*2,cBase+0 ,"  t0: ",0x10c);
	bT1 = WndEntry(rBase+rInc*2,cBase+21,"  t1: ",0x10c);
	bDt = WndEntry(rBase+rInc*2,cBase+42,"  dt: ",0x10c);

	WndGroup(nVport[0]-1,nVport[2],(nVport[3]-nVport[2])+1,
		(nVport[1]&0xff)-(nVport[0]&0xff)+1,"Preview",nSatrb);
	widVport = wndopensub(nVport,0);   // Open sub window for image
	wnd_select(widMain);
	rBase += 0xc00;
	cBase -= 1;
	bLcnt = WndEntry(rBase+rInc*3,cBase,szText[4]/*"Seg #: "*/,0x105);
	wnd_rloct(0,6); bType = wnd_field(0,0x7f,0x104);
	wnd_loct(rBase+rInc*3,cBase+19);
		bPreview= wnd_field(0,IsWin?0x1277:0x370,0x108);
		wnd_puts(szText[3]/*IsWin?"&Preview":" Preview"*/,0x270);
	wnd_loct(rBase+rInc*3,cBase+30);
		bOK = wnd_field(0,IsWin?0x1277:0x370,0x108);
		wnd_puts(szText[1]/*IsWin?"&Create":" Create "*/,0x270);
	wnd_loct(rBase+rInc*3,cBase+41);
		bCancel = wnd_field(0,IsWin?0x1277:0x370,0x108);
		wnd_puts(szText[2]/*IsWin?"&Quit":"  Quit  "*/,0x270);
	wnd_loct(rBase+rInc*3,cBase+52);
		bHelp = wnd_field(0,IsWin?0x1277:0x370,0x108);
		wnd_puts(szText[5]/*IsWin?"&Help":"  HELP  "*/,0x270);
}
//
//	CloseDialog() -- Close the main dialog window
//
void	CloseDialog()
{
	wndclose();
}
//
//	UpdateDialog() -- Update the content of the main dialog window
//
void	UpdateDialog()
{
	wnd_ftext(bXt,prints("%-54s",xt),0xf0);
	wnd_ftext(bYt,prints("%-54s",yt),0xf0);
	wnd_ftext(bT0,prints("%-12s",ts),0xf0);
	wnd_ftext(bT1,prints("%-12s",te),0xf0);
	wnd_ftext(bDt,prints("%-12s",dt),0xf0);
	if ( Lcnt )
		wnd_ftext(bLcnt,prints("%5ld",Lcnt),0xf0);
	else
		wnd_ftext(bLcnt,prints("%5s",""),0xf0);
	wnd_ftext(bType,prints("%-4s",szType[GenrType]),0x371);
}
//
int	WndEntry(int row, int col, char str[],int fsize)
{
	int	bID;

	wnd_loct(row,col);
	put3Dstr(str,nSatrb);
	bID = wnd_field(0,0x21f0,fsize);
	return bID;
}
//
//	WndGroup() -- Create Groupping window effect
//
void	WndGroup(int r,int c,int width, int height, char text[], int color)
{
	wnd_loct(r,c);
	wnd_rloct(height,0);
	if ( IsWin )
		wnd_field(0,0x277,(height<<8)|width);
	else
		wnd_field(0,((color&0xff)^0x08)|0x200,(height<<8)|width);
	wnd_itemno(1,-1);
	wnd_loct(r,c);
	if ( !IsWin )
		wnd_rloct(0xfc00,0);
	put3Dstr(text,color);
	wnd_loct(r,c);	      // Reset row offset, fix DOS version bug
}
//
//	Do we have direct support of some new TCL functions...
//
int	nSupport=identify("wnd_resize");
//
//	put3Dstr() -- Drawing 3D text
//
void	put3Dstr(char str[],int atrb)
{
	int	nLen;
	if ( nSupport ) {
		wnd_puts(str,atrb|0x300);
		return;
	}
	nLen=len(str);
	if ( nLen ) {
		wnd_rloct(0x100,IsWin?0x100:0);
		wnd_puts(str,(atrb^0x0f));
		wnd_rloct(0xff00,-nLen);
		wnd_puts(str,(atrb&0xff0f)|0x8000);
	}
}

char	szAsciiText[][40]={	    // Default English Display Text
/*  0 */   "Function Curve Setup Dialog",
/*  1 */   "&Create",
/*  2 */   "&Quit",
/*  3 */   "&Preview",
/*  4 */   " Seg#: ",
/*  5 */   "&Help",
};

char	szBig5Text[][36]={	   // Default Chinese Big-5 Display Text
/*  0 */   "uƹϧøs]w",
/*  1 */   "[&C]",
/*  2 */   "[&Q]",
/*  3 */   "w[&P]",
/*  4 */   " q: ",
/*  5 */   "[&H]",
};

char	szGB2312Text[][36]={	// Default Chinese GB2312 Display Text
/*  0 */   "ƽߺͼλ趨",
/*  1 */   "[&C]",
/*  2 */   "[&Q]",
/*  3 */   "Ԥ[&P]",
/*  4 */   " : ",
/*  5 */   "˵[&H]",
};
//
//	The following Japenese messages are translated by _____________.
//
char	szSJISText[][40]={    // Default Japenese SJIS Display Text
/*  0 */   "֐̐ݒ",
/*  1 */   "쐬[&C]",
/*  2 */   "I[&Q]",
/*  3 */   "\\[&P]",
/*  4 */   " Seg#: ",
/*  5 */   "[&H]",
};
//
//	The following Korean messages are translated by _____________.
//
char	szKoreanText[][40]={	// Default Korean Display Text
/*  0 */   "Function Curve Setup Dialog",
/*  1 */   "&Create",
/*  2 */   "&Quit",
/*  3 */   "&Preview",
/*  4 */   " Seg#: ",
/*  5 */   "&Help",
};

char	szDosAsciiText[][40]={	       // Default English Display Text
/*  0 */   "Function Curve Setup Dialog",
/*  1 */   " Create ",
/*  2 */   "  Quit  ",
/*  3 */   " Preview",
/*  4 */   " Seg#: ",
/*  5 */   "  Help  ",
};
char	szDosBig5Text[][26]={	      // Default Chinese Big-5 Display Text
/*  0 */   "uƹϧøs]w",
/*  1 */   "ͦu",
/*  2 */   "ާ@",
/*  3 */   "uw",
/*  4 */   " q: ",
/*  5 */   "    ",
};

char	szDosGB2312Text[][26]={    // Default Chinese GB2312 Display Text
/*  0 */   "ƽߺͼλ趨",
/*  1 */   "",
/*  2 */   "",
/*  3 */   "Ԥ",
/*  4 */   " : ",
/*  5 */   "  ˵  ",
};
//
//	The following Japenese messages are translated by _____________.
//
char	szDosSJISText[][40]={	 // Default Japenese SJIS Display Text
/*  0 */   "֐̐ݒ",
/*  1 */   "  쐬  ",
/*  2 */   "  I  ",
/*  3 */   "  \\  ",
/*  4 */   " Seg#: ",
/*  5 */   "    ",
};
//
//	The following Korean messages are translated by _____________.
//
char	szDosKoreanText[][40]={    // Default Korean Display Text
/*  0 */   "Function Curve Setup Dialog",
/*  1 */   " Create ",
/*  2 */   "  Quit  ",
/*  3 */   " Preview",
/*  4 */   " Seg#: ",
/*  5 */   "  Help  ",
};

#define BIG5		0x404
#define GB2312		0x804
#define SJIS		0x411
#define KOREAN		0x412
#define ASCII		0x409

#define TGBIG5		1
#define TGGB2312	2
#define TGSJIS		5
#define TGKOREAN	3

int	SetMsgText()
{
	char	szBuf[80];
	int	switched;

	IsWin = 1;
	switch(1) {  // Test if we are in TwinCAD ...
	  case !identify("os_type"):
	  case (os_type()<0):	// Must be in DOS version!
		IsWin = 0;
		break;
	}

	if ( IsWin ) {
	    szBuf = get_profile("General Setup","AlternateLanguage");
	    switched = 0;
	    if ( szBuf!="" )
		switched = eval(szBuf);
	    if ( switched ) {
		switch(os_language()) {
		  case BIG5:	setdata(szText,szBig5Text);	break;
		  case GB2312:	setdata(szText,szGB2312Text);	break;
		  case SJIS:	setdata(szText,szSJISText);	break;
		  case KOREAN:	setdata(szText,szKoreanText);	break;
		  default:	setdata(szText,szAsciiText);	break;
		}
	    }
	    else
		setdata(szText,szAsciiText);
	}
	else {	// Switched or not, depends on TGF type
		switch(@tgftype) {
		  case TGBIG5:	  setdata(szText,szDosBig5Text);     break;
		  case TGGB2312:  setdata(szText,szDosGB2312Text);   break;
		  case TGSJIS:	  setdata(szText,szDosSJISText);     break;
		  case TGKOREAN:  setdata(szText,szDosKoreanText);   break;
		  default:	setdata(szText,szDosAsciiText);    break;
		}
	}
	return 0;
}
